package intermediate

import (
	"context"
	"crypto/ecdsa"

	"git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/reputation/common"
	internalclient "git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/reputation/internal/client"
	coreclient "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/client"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/services/reputation"
	reputationcommon "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/services/reputation/common"
	eigentrustcalc "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/services/reputation/eigentrust/calculator"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util/logger"
	reputationapi "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/reputation"
	"go.uber.org/zap"
)

// RemoteProviderPrm groups the required parameters of the RemoteProvider's constructor.
//
// All values must comply with the requirements imposed on them.
// Passing incorrect parameter values will result in constructor
// failure (error or panic depending on the implementation).
type RemoteProviderPrm struct {
	Key *ecdsa.PrivateKey
	Log *logger.Logger
}

// NewRemoteProvider creates a new instance of the RemoteProvider.
//
// Panics if at least one value of the parameters is invalid.
//
// The created RemoteProvider does not require additional
// initialization and is completely ready for work.
func NewRemoteProvider(prm RemoteProviderPrm) *RemoteProvider {
	switch {
	case prm.Key == nil:
		common.PanicOnPrmValue("NetMapSource", prm.Key)
	case prm.Log == nil:
		common.PanicOnPrmValue("Logger", prm.Log)
	}

	return &RemoteProvider{
		key: prm.Key,
		log: prm.Log,
	}
}

// RemoteProvider is an implementation of the clientKeyRemoteProvider interface.
type RemoteProvider struct {
	key *ecdsa.PrivateKey
	log *logger.Logger
}

func (rp RemoteProvider) WithClient(c coreclient.Client) reputationcommon.WriterProvider {
	return &TrustWriterProvider{
		client: c,
		key:    rp.key,
		log:    rp.log,
	}
}

type TrustWriterProvider struct {
	client coreclient.Client
	key    *ecdsa.PrivateKey
	log    *logger.Logger
}

func (twp *TrustWriterProvider) InitWriter(ep reputationcommon.EpochProvider) (reputationcommon.Writer, error) {
	iterInfo, ok := ep.(eigentrustcalc.EpochIterationInfo)
	if !ok {
		// TODO: #1164 think if this can be done without such limitation
		panic(ErrIncorrectContextPanicMsg)
	}

	return &RemoteTrustWriter{
		iterInfo: iterInfo,
		client:   twp.client,
		key:      twp.key,
		log:      twp.log,
	}, nil
}

type RemoteTrustWriter struct {
	iterInfo eigentrustcalc.EpochIterationInfo
	client   coreclient.Client
	key      *ecdsa.PrivateKey
	log      *logger.Logger
}

// Write sends a trust value to a remote node via ReputationService.AnnounceIntermediateResult RPC.
func (rtp *RemoteTrustWriter) Write(ctx context.Context, t reputation.Trust) error {
	epoch := rtp.iterInfo.Epoch()
	i := rtp.iterInfo.I()

	rtp.log.Debug("announcing trust",
		zap.Uint64("epoch", epoch),
		zap.Uint32("iteration", i),
		zap.Stringer("trusting_peer", t.TrustingPeer()),
		zap.Stringer("trusted_peer", t.Peer()),
	)

	var apiTrust reputationapi.Trust
	apiTrust.SetValue(t.Value().Float64())
	apiTrust.SetPeer(t.Peer())

	var apiPeerToPeerTrust reputationapi.PeerToPeerTrust
	apiPeerToPeerTrust.SetTrustingPeer(t.TrustingPeer())
	apiPeerToPeerTrust.SetTrust(apiTrust)

	var p internalclient.AnnounceIntermediatePrm

	p.SetClient(rtp.client)
	p.SetEpoch(epoch)
	p.SetIteration(i)
	p.SetTrust(apiPeerToPeerTrust)

	_, err := internalclient.AnnounceIntermediate(ctx, p)

	return err
}

func (rtp *RemoteTrustWriter) Close(context.Context) error {
	return nil
}
