package allocator

import (
	"github.com/tidwall/btree"
)

// BestFit implements best-fit allocation strategy.
type BestFit struct {
	capacity   uint64
	sizeToFree *btree.BTreeG[Region]
	endToFree  *btree.BTreeG[Region]
}

func lessBySize(a, b Region) bool {
	if a.Length() == b.Length() {
		return a.Offset() < b.Offset()
	}
	return a.Length() < b.Length()
}

func lessByEnd(a, b Region) bool {
	return a.End() < b.End()
}

var _ Allocator = (*BestFit)(nil)

// NewBestFit returns new best-fit allocator instance.
func NewBestFit(capacity uint64) *BestFit {
	a := &BestFit{
		capacity:   capacity,
		sizeToFree: btree.NewBTreeGOptions(lessBySize, btree.Options{NoLocks: true}),
		endToFree:  btree.NewBTreeGOptions(lessByEnd, btree.Options{NoLocks: true}),
	}
	for i := uint64(0); i+maxLength < capacity; i += maxLength {
		a.addFreePortion(NewRegion(i*maxLength, maxLength))
	}
	c := capacity / maxLength * maxLength
	r := NewRegion(c, capacity-c)
	a.addFreePortion(NewRegion(0, 0))
	a.addFreePortion(r)
	return a
}

func seek(b *btree.BTreeG[Region], pivot Region) (Region, bool) {
	var item Region
	var seen bool
	b.Ascend(pivot, func(it Region) bool {
		item = it
		seen = true
		return false
	})
	return item, seen
}

// Get implements the Allocator interface.
func (b *BestFit) Get(size uint64) (Region, error) {
	p := NewRegion(0, uint64(size))

	free, ok := seek(b.sizeToFree, p)
	if ok {
		b.deleteFreePortion(free)

		reg := free.allocate(size)
		if free.Length() != 0 {
			b.addFreePortion(free)
		}
		return reg, nil
	}
	return 0, ErrOOM
}

// Free implements the Allocator interface.
func (b *BestFit) Free(region Region) error {
	p := b.mergeIfPossible(region)
	b.addFreePortion(p)
	return nil
}

func (b *BestFit) mergeIfPossible(free Region) Region {
	key := NewRegion(free.Offset(), 0)
	prev, ok := seek(b.endToFree, key)
	if ok {
		if prev.End() == free.Offset() && prev.safeExtend(free.Length()) {
			free = NewRegion(prev.Offset(), prev.Length()+free.Length())
			b.deleteFreePortion(prev)
		}
	}

	key = NewRegion(free.End(), 0)
	next, ok := seek(b.endToFree, key)
	if ok {
		if next.Offset() == free.End() && free.safeExtend(next.Length()) {
			free = NewRegion(free.Offset(), free.Length()+next.Length())
			b.deleteFreePortion(next)
		}
	}

	return free
}

func (b *BestFit) deleteFreePortion(p Region) {
	b.sizeToFree.Delete(p)
	b.endToFree.Delete(p)
}

func (b *BestFit) addFreePortion(p Region) {
	b.sizeToFree.Set(p)
	b.endToFree.Set(p)
}
