package allocator

import (
	"errors"
	"fmt"
	"math"
)

type Allocator interface {
	Get(uint64) (Region, error)
	Free(Region) error
}

var (
	ErrOOM = errors.New("no free chunk is found")
)

// Region describes continuous range of data as offset and length.
// Offset is 40-bit, for 512-byte chunks this means it can contain 2^40 * 2^9 = 512 TiB of data.
// Length is 24-bit, for 512-byte chunks this means it can contain 2^24 * 2^9 = 8 GiB of data.
type Region uint64

const (
	lenShift   = 40
	maxLength  = math.MaxUint64 >> lenShift
	offsetMask = (1 << lenShift) - 1
)

func NewRegion(offset, length uint64) Region {
	return Region((length << lenShift) | offset)
}

func (r Region) Offset() uint64 {
	return uint64(r & offsetMask)
}

func (r Region) Length() uint64 {
	return uint64(r >> lenShift)
}

func (r Region) End() uint64 {
	return r.Offset() + r.Length()
}

func (r *Region) safeExtend(size uint64) bool {
	return r.Length()+size < maxLength
}

func (r *Region) allocate(size uint64) Region {
	assert(size <= r.Length(), "invalid allocation size: have=%d, need=%d", r.Length(), size)

	result := NewRegion(r.Offset(), size)
	*r = NewRegion(r.Offset()+size, r.Length()-size)
	return result
}

func assert(ok bool, msg string, args ...any) {
	if !ok {
		panic(fmt.Sprintf(msg, args...))
	}
}
