package goodstor

import (
	"sync"

	"codeberg.org/fyrchik/uring/loop"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/allocator"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/compression"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/uringstor/slab"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util/logger"
	"go.uber.org/zap"
)

var _ common.Storage = (*Storage)(nil)

type Option = func(*Storage)

type Storage struct {
	capacity   uint64
	path       string
	blockSize  uint64
	noSync     bool
	readOnly   bool
	loopSize   int
	loopParams loop.Params
	loop       *loop.Loop
	log        *logger.Logger

	backend slab.Backend

	allocMtx  sync.Mutex
	allocator allocator.Allocator
}

const (
	defaultBlockSize     = 512
	defaultUringLoopSize = 1024
)

func New(opts ...Option) *Storage {
	var s Storage
	s.blockSize = defaultBlockSize
	s.loopSize = defaultUringLoopSize
	s.log = &logger.Logger{Logger: zap.NewNop()}
	for i := range opts {
		opts[i](&s)
	}

	s.allocator = allocator.NewBestFit(s.capacity / s.blockSize)
	return &s
}

// Type is uring storage type used in logs and configuration.
const Type = "goodstor"

// Type implements the common.Storage interface.
func (s *Storage) Type() string {
	return Type
}

// Path implements the common.Storage interface.
func (s *Storage) Path() string {
	return s.path
}

// SetCompressor implements the common.Storage interface.
func (s *Storage) SetCompressor(cc *compression.Config) {}

// SetReportErrorFunc allows to provide a function to be called on disk errors.
// This function MUST be called before Open.
func (s *Storage) SetReportErrorFunc(f func(string, error)) {}

// WithCapacity sets the max capacity of the storage.
func WithCapacity(capacity uint64) Option {
	return func(s *Storage) {
		s.capacity = capacity
	}
}

// WithRootPath sets the max capacity of the storage.
func WithRootPath(dir string) Option {
	return func(s *Storage) {
		s.path = dir
	}
}

func WithLoopParams(size int, p loop.Params) Option {
	return func(s *Storage) {
		s.loopSize = size
		s.loopParams = p
	}
}

func WithOSBackend() Option {
	return func(s *Storage) {
		s.loopSize = 0
	}
}

func WithBlockSize(size uint64) Option {
	return func(s *Storage) {
		s.blockSize = uint64(size)
	}
}

func WithLogger(l *logger.Logger) Option {
	return func(s *Storage) {
		s.log = &logger.Logger{Logger: l.With(zap.String("component", Type))}
	}
}

func WithNoSync(noSync bool) Option {
	return func(s *Storage) {
		s.noSync = noSync
	}
}
