package uringstor

import (
	"math/bits"
	"os"
	"path/filepath"
	"strconv"

	"codeberg.org/fyrchik/uring/loop"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/uringstor/slab"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util"
)

const idToStringBase = 16

type nameIDPair struct {
	name string
	id   uint16
}

func (s *Storage) Open(readOnly bool) error {
	s.readOnly = readOnly

	if s.loopSize != 0 {
		lp, err := loop.New(s.loopSize, &s.loopParams)
		if err != nil {
			return err
		}
		s.loop = lp
	}

	if !readOnly {
		if err := util.MkdirAllX(s.path, os.ModePerm); err != nil {
			return err
		}
	}

	names, err := os.ReadDir(s.path)
	if err != nil {
		return err
	}

	ids := make([]nameIDPair, 0, len(names))
	for i := range names {
		id, err := strconv.ParseUint(names[i].Name(), idToStringBase, 16)
		if err != nil {
			return err
		}
		if _, ok := s.slabMap[uint16(id)]; ok {
			continue
		}
		ids = append(ids, nameIDPair{
			name: names[i].Name(),
			id:   uint16(id),
		})
	}

	maxID := -1
	for i := range ids {
		slabPath := filepath.Join(s.path, ids[i].name)
		sl, err := s.openSlab(slabPath, 0, 0)
		if err != nil {
			return err
		}

		slabID := ids[i].id
		index := bits.TrailingZeros32(sl.ChunkSize()) - bits.TrailingZeros32(uint32(s.minObjectSize))
		s.slabIDs[index] = append(s.slabIDs[index], slabID)
		s.slabMap[slabID] = sl

		if int(slabID) > maxID {
			maxID = int(slabID)
		}
	}

	if maxID > 0 {
		s.nextID = uint16(maxID + 1)
	}
	for i := range s.slabIDs {
		if len(s.slabIDs[i]) != 0 {
			continue
		}

		slabPath := filepath.Join(s.path, strconv.FormatUint(uint64(s.nextID), idToStringBase))
		ss, err := s.openSlab(slabPath, uint32(s.minObjectSize<<i), s.slabCapacity)
		if err != nil {
			return err
		}

		s.slabIDs[i] = append(s.slabIDs[i], s.nextID)
		s.slabMap[s.nextID] = ss
		s.nextID++
	}
	return nil
}

func (s *Storage) Init() error {
	return nil
}

func (s *Storage) Close() error {
	for _, sl := range s.slabMap {
		if err := sl.Close(); err != nil {
			return err
		}
	}

	s.nextID = 0
	s.slabMap = make(map[uint16]*slab.Slab)
	for i := range s.slabIDs {
		s.slabIDs[i] = nil
	}

	if s.loop != nil {
		s.loop.Close()
		s.loop = nil
	}
	return nil
}
