package uringstor

import (
	"math/bits"
	"path/filepath"
	"strconv"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/uringstor/slab"
	storagelog "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/internal/log"
	"go.uber.org/zap"
)

func (s *Storage) Put(prm common.PutPrm) (common.PutRes, error) {
	if s.readOnly {
		return common.PutRes{}, common.ErrReadOnly
	}
	if prm.RawData == nil {
		panic("unexpected")
	}

	up2 := s.minObjectSize
	for ; up2 < 5+uint64(len(prm.RawData)); up2 *= 2 {
	}

	bucket := bits.TrailingZeros64(up2) - bits.TrailingZeros64(s.minObjectSize)

	s.slabMtx.RLock()
	ln := len(s.slabIDs[bucket])
	for i := 0; i < ln; i++ {
		slabID := s.slabIDs[bucket][i]
		ss, ok := s.slabMap[slabID]
		assert(ok, "slab ID is here, but the slab is missing: %d", slabID)
		s.slabMtx.RUnlock()

		//start := time.Now()
		offset, err := ss.Put(prm.RawData)
		//fmt.Println(time.Since(start))
		if err == nil {
			storageID := marshalStorageID(slabID, offset)
			// storagelog.Write(s.log,
			// 	storagelog.OpField("PUT"),
			// 	storagelog.StorageIDField(storageID[:]),
			// 	storagelog.AddressField(prm.Address))
			return common.PutRes{StorageID: storageID[:]}, nil
		}
		s.log.Error("can't put object to a slab", zap.Error(err))

		s.slabMtx.RLock()
	}
	s.slabMtx.RUnlock()

	s.slabMtx.Lock()
	var ss *slab.Slab
	var slabID uint16
	if ln < len(s.slabIDs[bucket]) {
		// New slab was created by other thread.
		// FIXME (@fyrchik): In theory it can already be filled, in practice it is only noticeable on big object sizes.
		slabID = s.slabIDs[bucket][ln]
		ss = s.slabMap[slabID]
	} else {
		var err error
		slabPath := filepath.Join(s.path, strconv.FormatUint(uint64(s.nextID), idToStringBase))
		ss, err = s.openSlab(slabPath, uint32(up2), s.slabCapacity)
		if err != nil {
			s.slabMtx.Unlock()
			return common.PutRes{}, err
		}

		slabID = s.nextID
		s.slabIDs[bucket] = append(s.slabIDs[bucket], slabID)
		s.slabMap[slabID] = ss
		s.nextID++

		s.log.Debug("opened new slab",
			zap.Uint16("slab_id", slabID),
			zap.Uint32("chunk_size", uint32(up2)),
			zap.Uint64("capacity", s.slabCapacity))
	}
	s.slabMtx.Unlock()

	offset, err := ss.Put(prm.RawData)
	if err != nil {
		return common.PutRes{}, nil
	}

	storageID := marshalStorageID(slabID, offset)
	storagelog.Write(s.log,
		storagelog.OpField("PUT"),
		storagelog.StorageIDField(storageID[:]),
		storagelog.AddressField(prm.Address))
	return common.PutRes{StorageID: storageID[:]}, nil
}
