package slab

import (
	"fmt"
	"os"

	"codeberg.org/fyrchik/uring"
	"codeberg.org/fyrchik/uring/loop"
	"golang.org/x/sys/unix"
)

// Backend
type Backend interface {
	Open(flags int) error
	Close() error
	Fallocate(size int64) error

	WriteAt(buf []byte, offset int64) error
	ReadAt(buf []byte, offset int64) error
}

type File struct {
	path string
	file *os.File
}

type UringFile struct {
	File
	loop *loop.Loop
	fd   uintptr
}

// Allow to read unwritten but preallocated portions of file.
// Because of this, `t.Name()` should be used for directories, not `t.TempDir()` in tests.
const _FALLOC_FL_ZERO_RANGE = 16

func NewFileBackend(p string) *File {
	return &File{path: p}
}

func (f *File) Open(flags int) error {
	fi, err := os.OpenFile(f.path, flags|os.O_CREATE|os.O_SYNC, os.ModePerm)
	if err != nil {
		return err
	}

	f.file = fi
	return nil
}

func (f *File) Close() error {
	if f.file == nil {
		return nil
	}

	if err := f.file.Close(); err != nil {
		return err
	}

	f.file = nil
	return nil
}

func (f *File) Fallocate(size int64) error {
	return unix.Fallocate(int(f.file.Fd()), _FALLOC_FL_ZERO_RANGE, 0, size)
}

func (f *File) WriteAt(buf []byte, offset int64) error {
	_, err := f.file.WriteAt(buf, offset)
	if err != nil {
		return fmt.Errorf("os: %w", err)
	}
	return nil
}

func (f *File) ReadAt(buf []byte, offset int64) error {
	_, err := f.file.ReadAt(buf, offset)
	return err
}

func NewUringBackend(p string, l *loop.Loop) *UringFile {
	return &UringFile{
		File: File{
			path: p,
		},
		loop: l,
	}
}

func (f *UringFile) Open(flags int) error {
	if err := f.File.Open(flags); err != nil {
		return err
	}

	fd := f.File.file.Fd()
	//if err := f.loop.RegisterFiles([]int32{int32(fd)}); err != nil {
	//	return err
	//}
	f.fd = fd
	return nil
}

func (f *UringFile) Close() error {
	if err := f.File.Close(); err != nil {
		return err
	}

	f.fd = 0
	return nil
}

func (f *UringFile) Fallocate(size int64) error {
	var sqe uring.SQEntry
	uring.Fallocate(&sqe, f.fd, _FALLOC_FL_ZERO_RANGE, 0, uint64(size))
	//sqe.SetFlags(uring.IOSQE_FIXED_FILE)

	_, err := f.loop.Complete(sqe)
	return err
}

func (f *UringFile) WriteAt(buf []byte, offset int64) error {
	var sqe uring.SQEntry
	uring.WriteAt(&sqe, f.fd, buf, uint64(offset))
	//sqe.SetFlags(uring.IOSQE_FIXED_FILE)

	_, err := f.loop.Complete(sqe)
	if err != nil {
		return fmt.Errorf("uring: %w", err)
	}
	return err
}

func (f *UringFile) ReadAt(buf []byte, offset int64) error {
	var sqe uring.SQEntry
	uring.ReadAt(&sqe, f.fd, buf, uint64(offset))
	//sqe.SetFlags(uring.IOSQE_FIXED_FILE)

	_, err := f.loop.Complete(sqe)
	return err
}
