package slab

import (
	"encoding/binary"
	"errors"
	"math/bits"
	"os"
	"sync"

	"github.com/RoaringBitmap/roaring"
)

const (
	minChunkSize = 1024
)

type header struct {
	chunkSize uint32
	size      uint64
}

type Slab struct {
	header
	backend Backend

	mtx    sync.Mutex
	bitmap *roaring.Bitmap
}

const (
	maxBitmapSize = 1024 * 1024
	headerSize    = 4 /* runtime state */ + 4 /* chunk size */ + 8 /* total size */
)

const (
	headerUnsynced = iota
	headerSynced
)

type Params struct {
	Backend   Backend
	ChunkSize uint32
	Capacity  uint64
	ReadOnly  bool
	NoSync    bool
}

func (p Params) openFlags() int {
	flags := os.O_RDWR
	if p.ReadOnly {
		flags = os.O_RDONLY
	}
	if !p.NoSync {
		flags |= os.O_SYNC
	}
	return flags
}

// New opens ans -returns new slab.
// If chunkSize and size are zero, a slab must exist and all entries are read from the slab header.
// Otherwise, both of them must be a positive power of 2 and a new file is created.
func New(p Params) (*Slab, error) {
	s := &Slab{
		backend: p.Backend,
		header: header{
			chunkSize: p.ChunkSize,
			size:      p.Capacity,
		},
	}

	if p.ChunkSize == 0 && p.Capacity == 0 {
		return s, s.open(p.openFlags())
	}
	if p.ReadOnly {
		return nil, errors.New("slab: creation is not supported in read-only mode")
	}
	if p.ChunkSize < minChunkSize {
		return nil, errors.New("slab: chunk size too small")
	}
	if bits.OnesCount32(p.ChunkSize) != 1 {
		return nil, errors.New("slab: chunk size must be a power of 2")
	}
	if bits.OnesCount64(p.Capacity) != 1 {
		return nil, errors.New("slab: size must be a power of two")
	}
	return s, s.create()
}

// ChunkSize returns chunk size for this slab.
func (s *Slab) ChunkSize() uint32 {
	return s.chunkSize
}

// Size returns size of this slab minus header size.
func (s *Slab) Size() uint64 {
	return s.size
}

func (s *Slab) create() error {
	if err := s.backend.Open(os.O_RDWR | os.O_EXCL); err != nil {
		return err
	}

	if err := s.backend.Fallocate(int64(s.size)); err != nil {
		return err
	}

	s.bitmap = roaring.NewBitmap()
	s.bitmap.AddRange(0, s.size/uint64(s.chunkSize))
	if err := s.writeHeader(); err != nil {
		return err
	}
	return s.writeSyncFlag(headerUnsynced)
}

func (s *Slab) open(flags int) error {
	if err := s.backend.Open(flags); err != nil {
		return err
	}
	if err := s.readHeader(); err != nil {
		return err
	}
	if flags&0x3 == os.O_RDONLY { // FIXME (@fyrchik): too tricky, rewrite.
		return nil
	}
	return s.writeSyncFlag(headerUnsynced)
}

func (s *Slab) readHeader() error {
	buf := make([]byte, headerSize)
	if err := s.backend.ReadAt(buf, 0); err != nil {
		return err
	}
	if binary.LittleEndian.Uint32(buf) != headerSynced {
		return errors.New("slab was not closed properly, reiteration is needed")
	}

	chunkSize := binary.LittleEndian.Uint32(buf[4:])
	size := binary.LittleEndian.Uint64(buf[8:])

	bitmapSize := size / uint64(chunkSize)
	buf = make([]byte, bitmapSize/8)
	if err := s.backend.ReadAt(buf, headerSize); err != nil {
		return err
	}

	bitmap := roaring.New()
	for i := 0; i < len(buf); i++ {
		for j := 0; j < 8; j++ {
			if buf[i]&(1<<j) != 0 {
				bitmap.Add(uint32(i*8 + 7 - j))
			}
		}
	}

	s.chunkSize = chunkSize
	s.size = size
	s.bitmap = bitmap
	return nil
}

func (s *Slab) headerSize() uint64 {
	return headerSize + s.size/uint64(s.chunkSize)/8
}

func (s *Slab) writeHeader() error {
	buf := make([]byte, s.headerSize())
	binary.LittleEndian.PutUint32(buf, headerSynced)
	binary.LittleEndian.PutUint32(buf[4:], s.chunkSize)
	binary.LittleEndian.PutUint64(buf[8:], s.size)

	bmBuf := buf[headerSize:]
	for i := 0; i < len(bmBuf); i++ {
		for j := 0; j < 8; j++ {
			if s.bitmap.Contains(uint32(i*8 + 7 - j)) {
				bmBuf[i] |= 1 << j
			}
		}
	}

	return s.backend.WriteAt(buf, 0)
}

// Close closes slab.
func (s *Slab) Close() error {
	s.writeSyncFlag(headerSynced)
	return s.backend.Close()
}

func (s *Slab) writeSyncFlag(flag uint32) error {
	buf := make([]byte, 4)
	binary.LittleEndian.PutUint32(buf, flag)
	return s.backend.WriteAt(buf, 0)
}

const (
	empty = iota
	exclusive
)

const hdrSize = 5

func (s *Slab) Put(value []byte) (uint64, error) {
	if 5+len(value) > int(s.chunkSize) {
		return 0, errors.New("boundary")
	}

	s.mtx.Lock()
	if s.bitmap.GetCardinality() == 0 {
		s.mtx.Unlock()
		return 0, errors.New("no space")
	}

	slot := s.bitmap.Minimum()
	s.bitmap.Remove(slot)
	s.mtx.Unlock()

	buf := make([]byte, 5+len(value))
	buf[0] = exclusive
	binary.LittleEndian.PutUint32(buf[1:], uint32(len(value)))
	copy(buf[5:], value)

	offset := uint64(slot * s.chunkSize)
	err := s.backend.WriteAt(buf, int64(offset+s.headerSize()))
	if err != nil {
		s.mtx.Lock()
		s.bitmap.Add(slot)
		s.mtx.Unlock()
	}
	return uint64(offset), err
}

func (s *Slab) Get(offset uint64) ([]byte, error) {
	if offset%uint64(s.chunkSize) != 0 {
		return nil, errors.New("slab: invalid offset alignment")
	}
	if offset >= s.size {
		return nil, errors.New("slab: offset too big")
	}

	buf := make([]byte, s.chunkSize)
	err := s.backend.ReadAt(buf, int64(offset+s.headerSize()))
	if buf[0] != exclusive {
		return nil, errors.New("not found")
	}

	ln := binary.LittleEndian.Uint32(buf[1:])
	if ln+5 > uint32(len(buf)) {
		return nil, errors.New("corrupted data")
	}
	return buf[5 : 5+int(ln)], err
}

func (s *Slab) Exists(offset uint64) (bool, error) {
	buf := make([]byte, 1)
	err := s.backend.ReadAt(buf, int64(offset+s.headerSize()))
	if err != nil {
		return false, err
	}

	return buf[0] == exclusive, nil
}

var emptyMark = []byte{empty}

func (s *Slab) Delete(offset uint64) error {
	if offset%uint64(s.chunkSize) != 0 {
		return errors.New("slab: invalid offset alignment")
	}
	if offset >= s.size {
		return errors.New("slab: offset too big")
	}

	err := s.backend.WriteAt(emptyMark, int64(offset+s.headerSize()))
	if err != nil {
		return err
	}

	s.mtx.Lock()
	s.bitmap.Remove(uint32(offset / uint64(s.chunkSize)))
	s.mtx.Unlock()
	return nil
}
