package slab

import (
	"crypto/rand"
	"os"
	"path/filepath"
	"testing"

	"codeberg.org/fyrchik/uring/loop"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util"
	"github.com/stretchr/testify/require"
)

func TestSlab(t *testing.T) {
	t.Run("file", func(t *testing.T) {
		path := filepath.Join(t.Name(), "slab")
		_ = os.RemoveAll(path)
		defer os.RemoveAll(path)

		require.NoError(t, util.MkdirAllX(t.Name(), os.ModePerm))
		b := NewFileBackend(path)
		testSlab(t, b)
	})
	t.Run("uring", func(t *testing.T) {
		l, err := loop.New(10, &loop.Params{RingCount: 1})
		require.NoError(t, err)
		defer l.Close()

		path := filepath.Join(t.Name(), "slab")
		_ = os.RemoveAll(path)
		defer os.RemoveAll(path)

		require.NoError(t, util.MkdirAllX(t.Name(), os.ModePerm))
		b := NewUringBackend(path, l)
		testSlab(t, b)
	})

}

func testSlab(t *testing.T, b Backend) {
	objects := []struct {
		data   []byte
		offset int64
	}{
		{randomBytes(1), -1},
		{randomBytes(2), -1},
		{randomBytes(3), -1},
	}

	s, err := New(Params{
		Backend:   b,
		ChunkSize: 1024,
		Capacity:  1024 * 1024,
	})
	require.NoError(t, err)
	defer func() { s.Close() }()

	for i := range objects {
		id, err := s.Put(objects[i].data)
		require.NoError(t, err)

		objects[i].offset = int64(id)
	}

	for i := range objects {
		actual, err := s.Get(uint64(objects[i].offset))
		require.NoError(t, err)
		require.Equal(t, objects[i].data, actual)
	}

	require.NoError(t, s.Delete(uint64(objects[0].offset)))
	require.NoError(t, s.Delete(uint64(objects[2].offset)))

	_, err = s.Get(uint64(objects[0].offset))
	require.Error(t, err)

	_, err = s.Get(uint64(objects[2].offset))
	require.Error(t, err)

	actual, err := s.Get(uint64(objects[1].offset))
	require.NoError(t, err)
	require.Equal(t, objects[1].data, actual)

	require.NoError(t, s.Close())

	s, err = New(Params{Backend: b})
	require.NoError(t, err)

	_, err = s.Get(uint64(objects[0].offset))
	require.Error(t, err)

	_, err = s.Get(uint64(objects[2].offset))
	require.Error(t, err)

	actual, err = s.Get(uint64(objects[1].offset))
	require.NoError(t, err)
	require.Equal(t, objects[1].data, actual)

}

func randomBytes(n int) []byte {
	a := make([]byte, n)
	rand.Read(a)
	return a
}
