package uringstor

import (
	"math/bits"
	"sync"

	"codeberg.org/fyrchik/uring/loop"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/compression"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/uringstor/slab"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/util/logger"
	"go.uber.org/zap"
)

var _ common.Storage = (*Storage)(nil)

type Option = func(*Storage)

type Storage struct {
	slabCapacity  uint64
	path          string
	maxObjectSize uint64
	minObjectSize uint64
	noSync        bool
	readOnly      bool
	loopSize      int
	loopParams    loop.Params
	loop          *loop.Loop
	log           *logger.Logger

	// slabMtx protects slabs, slabIDs and nextID fields.
	slabMtx sync.RWMutex
	// ID of the next slab to create.
	nextID uint16
	// slabMap maps slab ID to a slab instance.
	slabMap map[uint16]*slab.Slab
	// slabIDs contains slices of used slabs sorted by size.
	// Group at index i contains slabs storing objects of size <= (minObjectSize << i).
	// The size of this array is fixed at the creation time.
	// The size of elements can only increase.
	slabIDs [][]uint16
}

const (
	defaultMaxObjectSize = 256 * 1024 * 1024
	defaultMinObjectSize = 1024
	defaultUringLoopSize = 1024
)

func New(opts ...Option) *Storage {
	var s Storage
	s.maxObjectSize = defaultMaxObjectSize
	s.minObjectSize = defaultMinObjectSize
	s.loopSize = defaultUringLoopSize
	s.log = &logger.Logger{Logger: zap.NewNop()}
	for i := range opts {
		opts[i](&s)
	}

	count := bits.TrailingZeros64(s.maxObjectSize) - bits.TrailingZeros64(s.minObjectSize) + 1
	s.slabMap = make(map[uint16]*slab.Slab)
	s.slabIDs = make([][]uint16, count)
	return &s
}

// Type is uring storage type used in logs and configuration.
const Type = "uringstor"

// Type implements the common.Storage interface.
func (s *Storage) Type() string {
	return Type
}

// Path implements the common.Storage interface.
func (s *Storage) Path() string {
	return s.path
}

// SetCompressor implements the common.Storage interface.
func (s *Storage) SetCompressor(cc *compression.Config) {}

// SetReportErrorFunc allows to provide a function to be called on disk errors.
// This function MUST be called before Open.
func (s *Storage) SetReportErrorFunc(f func(string, error)) {}

// WithSlabCapacity sets the max capacity of the storage.
func WithSlabCapacity(capacity uint64) Option {
	return func(s *Storage) {
		s.slabCapacity = capacity
	}
}

// WithRootPath sets the max capacity of the storage.
func WithRootPath(dir string) Option {
	return func(s *Storage) {
		s.path = dir
	}
}

func WithLoopParams(size int, p loop.Params) Option {
	return func(s *Storage) {
		s.loopSize = size
		s.loopParams = p
	}
}

func WithOSBackend() Option {
	return func(s *Storage) {
		s.loopSize = 0
	}
}

func WithMaxObjectSize(size int64) Option {
	return func(s *Storage) {
		s.maxObjectSize = uint64(size)
	}
}

func WithLogger(l *logger.Logger) Option {
	return func(s *Storage) {
		s.log = &logger.Logger{Logger: l.With(zap.String("component", Type))}
	}
}

func WithNoSync(noSync bool) Option {
	return func(s *Storage) {
		s.noSync = noSync
	}
}
